#include "_cl.h"

#include "method.h"
#include "stdio.h"
#include "files.h"
// implement buffered output.

#define MAX_BUFF 10240
#define MAX_FILES_INCR 1024

static SFILE **SFiles;
static int nMaxSFiles;
int nsfiles_array_pos_reached=0;
int ntotal_files_opened;

static void CompactifySFilesArray(void)
{
 int i=0;
 int ipos=0;
 for(i=0;i<nsfiles_array_pos_reached;i++)
  if(SFiles[i])
  {
   if(ipos==i)
    ipos++;
   else
   {
    SFILE *sf=SFiles[i];
    sf->nrecord=i;
    SFiles[ipos++]=SFiles[i];
   }
  }
 nsfiles_array_pos_reached=ipos;
}

static void MingRecordNewSFile(SFILE *sf)
{
 int i=0;
 for(i=0;i<nsfiles_array_pos_reached;i++)
  if(SFiles[i]==NULL)
  {
   SFiles[i]=sf;
   return;
  }
 if(nsfiles_array_pos_reached+1>nMaxSFiles)
 {
  int new_nMaxSFiles=nMaxSFiles+MAX_FILES_INCR;
  SFILE **NewSFiles=(SFILE**)_realloc_((void *)SFiles,new_nMaxSFiles*sizeof(SFILE *));
  if(NewSFiles)
  {
   nMaxSFiles=new_nMaxSFiles;
   SFiles=NewSFiles;
  }
  else
  {
   if(NewSFiles)_free_(NewSFiles);
   return;
  }
 }
 sf->nrecord=nsfiles_array_pos_reached;
 SFiles[nsfiles_array_pos_reached++]=sf;
 ntotal_files_opened++;
 if(ntotal_files_opened%1000==0)
  CompactifySFilesArray();
 return;
}

static void MingUnRecordSFile(SFILE *sf)
{
 int i=sf->nrecord;
 if(i>=0&&i<nsfiles_array_pos_reached)
  if(SFiles[i]==sf)
  {
   SFiles[i]=NULL;
   return;
  }
}

void show_last_error(void)
{
 char szmsg[1024];
 sprintf(szmsg,"File output error: %s",_strerror(NULL));
 SWF_error(szmsg);
}



WM_DLL_A void WM_DLL_B MingCloseAllFiles(void)
{
 int i=0;
 for(i=0;i<nsfiles_array_pos_reached;i++)
  if(SFiles[i])
  {
   fclose_Buf(SFiles[i]);
  }
 if(SFiles)_free_(SFiles);
 nsfiles_array_pos_reached=nMaxSFiles=0;
 SFiles=NULL;
 return;
}

SWFByteOutputMethod SWFByteOutput_User;
SWFFileOpenMethod SWFFileOpen_User;
SWFFileCloseMethod SWFFileClose_User;

WM_DLL_A void WM_DLL_B Set_fileByteOutputMethod(SWFByteOutputMethod new_method)
{
 SWFByteOutput_User=new_method;
}

WM_DLL_A void WM_DLL_B Set_fileOpenMethod(SWFFileOpenMethod new_method)
{
 SWFFileOpen_User=new_method;
}

WM_DLL_A void WM_DLL_B Set_fileCloseMethod(SWFFileCloseMethod new_method)
{
 SWFFileClose_User=new_method;
}

WM_DLL_A void WM_DLL_B fclose_Buf(SFILE *sf)
{
 int i=0;
 if(SWFFileClose_User)
  SWFFileClose_User(sf);
 else
 if(sf)
 {
  FILE *f=sf->f;
  if(f)
  {
   if(sf->ipos>0)
    fwrite(sf->Buff,1,sf->ipos,f);
   fclose(f);
   MingUnRecordSFile(sf);
   _free_(sf);
  }
  if(ferror(f))
  {
   char szmsg[1024];
   sprintf(szmsg,"Couldn't close %s to save data:\n\nError message from C run time library:\n%s"
    ,sf->szFile, strerror(errno));
   SWF_warn(szmsg);
  }
 }
}

WM_DLL_A SFILE * WM_DLL_B fopen_Buf(char *szFile,const char *mode)
{
 SFILE *sf=NULL;
 FILE *f=NULL;
 if(SWFFileOpen_User)
  return SWFFileOpen_User(szFile,mode);
 f=fopen(szFile,mode);
 if(f)
 {
  sf=(SFILE *)_malloc_(sizeof(SFILE));
  if(sf)
  {
   MingRecordNewSFile(sf);
   sf->f=f;
   sf->ipos=0;
   strncpy(sf->szFile,szFile,SFILE_MAX_FILE_NAME);
   sf->szFile[SFILE_MAX_FILE_NAME-1]='\0';
  }
 }
 else
 {
  char szmsg[1024];
  sprintf(szmsg,"Couldn't open %s to save data:\n\nError message from C run time library:\n%s"
   ,szFile, strerror(errno));
  SWF_warn(szmsg);
 }
 return sf;
}


void fileOutputMethod(byte b,void *data)
{
 SFILE *sf=(SFILE *)data;
 if(SWFByteOutput_User)
 {
  SWFByteOutput_User(b,data);
  return;
 }
 if(sf)
 {
  FILE *f=sf->f;
  char *Buff=sf->Buff;
  if(f)
  {
   SWF_assert(sf->ipos>=0);
   SWF_assert(sf->ipos<SFILE_BUFF_SIZE);
   if(sf->ipos<SFILE_BUFF_SIZE)
   {
    Buff[sf->ipos]=b;
    sf->ipos++;
   }
   if(sf->ipos>=SFILE_BUFF_SIZE)
   {
    fwrite(Buff,1,sf->ipos,f);
    sf->ipos=0;
   }
   if(ferror(f))
   {
    char szmsg[1024];
    sprintf(szmsg,"Error saving to %s\n\nError message from C run time library:\n%s"
     ,sf->szFile, strerror(errno));
    SWF_warn(szmsg);
    sf->f=NULL;// to prevent further output
   }
  }
  if(ferror(f))
  {
   char szmsg[1024];
   sprintf(szmsg,"Error saving to %s\n\nError message from C run time library:\n%s"
    ,sf->szFile, strerror(errno));
   SWF_warn(szmsg);
  }
 }
 return;
}


  
                 